@@ -70,13 +70,15 @@ class LuaScripts(object):
         KEYS[6]: the key for new user detection
         KEYS[7]: the key for the stitch pool
         KEYS[8]: temporary key to do some operations
+        KEYS[9]: all stitches key
+        KEYS[10]: all stitch pools key
         ARGV[1]: profile update info (ARGV[3])
         ARGV[2]: current timestamp (ARGV[4])
         ARGV[3]: session id (ARGV[5])
         ARGV[4]: user id (ARGV[6])
         ARGV[5]: array with pointers (can be empty) (ARGV[7])
         ARGV[6]: array with stitch keys (can be empty)
-        ARGV[7]: max lenghth stitch objects (pool & stitch itself)
+        ARGV[7]: max length stitch objects (pool & stitch itself)
         '''
         return '''
             local new_user = (redis.call('EXISTS', KEYS[1]) == 0)
@@ -133,6 +135,7 @@ class LuaScripts(object):
             local stitch_keys = cjson.decode(ARGV[6])
             if #stitch_keys > 0 then
                 local max_len_stitch_objects = tonumber(ARGV[7])
+                local stitch_keys_with_ts = {{}}
                 for idx, el in pairs(stitch_keys) do
                     redis.call('ZADD', el, timestamp_now, ARGV[4])
                     local len_stitch = redis.call('ZCARD', el)
@@ -140,15 +143,15 @@ class LuaScripts(object):
                         redis.call('ZREMRANGEBYRANK', el, 0, len_stitch - max_len_stitch_objects - 1) 
                     end
                     redis.call('EXPIRE', el, {user_ttl_with_margin})
-                    redis.call('ZADD', KEYS[7], timestamp_now, el)
                 end
-                if #stitch_keys > 0 then
-                    local len_stitch_pool = redis.call('ZCARD', KEYS[7])                    
-                    if len_stitch_pool > max_len_stitch_objects then
-                        redis.call('ZREMRANGEBYRANK', KEYS[7], 0, len_stitch_pool - max_len_stitch_objects - 1)                    
-                    end
+                redis.call('ZADD', KEYS[7], unpack(stitch_keys_with_ts))
+                redis.call('ZADD', KEYS[9], unpack(stitch_keys_with_ts)) -- this one does not expire
+                local len_stitch_pool = redis.call('ZCARD', KEYS[7])                    
+                if len_stitch_pool > max_len_stitch_objects then
+                    redis.call('ZREMRANGEBYRANK', KEYS[7], 0, len_stitch_pool - max_len_stitch_objects - 1)                    
                 end
                 redis.call('EXPIRE', KEYS[7], {user_ttl_with_margin})
+                redis.call('ZADD', KEYS[10], timestamp_now, KEYS[7])
             end
 
             -- add user_hash_key, created, and updated to profile
@@ -518,7 +521,7 @@ class LuaScripts(object):
                 for idx, el in pairs(stitch_keys) do
 
                     local stitch_key_result = {}
-                    local user_ids = redis.call('ZRANGE', el, 0, -1)
+                    local user_ids = redis.call('ZRANGE', el, 0, -1, 'WITHSCORES')
                     table.insert(stitch_key_result, 'user_ids')
                     table.insert(stitch_key_result, user_ids)
 
