@@ -625,7 +625,9 @@ class Profile(object):
                 self.profile_pool_key,
                 self.new_user_detected_key,
                 self.stitch_pool_key,
-                self.temp_key],
+                self.temp_key,
+                self.all_stitches_keys,
+                self.all_stitch_pools_key],
             args=[
                 self.lua_update_string,
                 self.time.timestamp(),
@@ -649,6 +651,16 @@ class Profile(object):
         # this is also defined on the stitch pool class!
         return f'st:u:p:{self.domain}:{self.user_id}'
 
+    @property
+    def all_stitch_pools_key(self):
+        # this is also defined on the stitch pool class!
+        return f'st:u:p:all:{self.domain}'
+
+    @property
+    def all_stitches_keys(self):
+        # this is also defined on the stitch class!
+        return f'st:u:all:{self.domain}'
+
     def create_pointers(self, pointers):
 
         if pointers is None:
