@@ -1,6 +1,8 @@
 import json
 import datetime
 import pytz
+import time
+import random
 
 DEFAULT_DOMAIN = None
 DEFAULT_REDIS_CLIENT = None
@@ -45,6 +47,9 @@ class StitchPool(object):
         assert isinstance(stitch_keys, (list, set)), 'stitch_keys needs to be a list or set'
         self.stitch_keys = self.stitch_keys.union(set(stitch_keys))
 
+    def delete_from_cache(self):
+        self.redis_client.delete(self.key)
+
     def to_cache(self):
 
         now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
@@ -62,17 +67,28 @@ class StitchPool(object):
 
     def complement_from_db(self):
 
+        model = self.usable_django_model()
         try:
-            stitch_pool_model = self.django_stitch_pool_model.objects.get(user_id=self.user_id)
+            stitch_pool_model = model.objects.get(user_id=self.user_id)
             stitches_keys = [str(s) for s in stitch_pool_model.stitch_keys]
             self.add_stitch_keys(stitches_keys)
             if len(stitches_keys) > 0:
                 self.to_cache()
             self.set_recent_db_check()
-        except self.django_stitch_pool_model.DoesNotExist:
+        except model.DoesNotExist:
             self.set_recent_db_check()
             return []
 
+    @classmethod
+    def usable_django_model(cls):
+        if cls.django_stitch_pool_model:
+            if all(cls.django_connection_tuple):
+                if cls.django_connection_tuple[0].connection and not cls.django_connection_tuple[0].is_usable():
+                    del cls.django_connection_tuple[1]._connections.default
+                return cls.django_stitch_pool_model
+            else:
+                return cls.django_stitch_pool_model
+
     @property
     def recently_checked_db(self):
         if self._recently_checked_db is None:
@@ -101,6 +117,14 @@ class StitchPool(object):
     def set_recent_db_check(self):
         self.redis_client.set(self.recently_checked_db_key, "1", ex=self.user_ttl)
 
+    @property
+    def all_stitchpools_key(self):
+        return self._all_stitchpools_key()
+
+    @classmethod
+    def _all_stitchpools_key(cls):
+        return f'st:u:p:all:{cls.domain}'
+
 class Stitch(object):
 
     default_type = 'default'
@@ -117,15 +141,17 @@ class Stitch(object):
                  value,
                  type,
                  user_ids=None,
+                 timestamps=None,
                  recently_checked_db=None):
 
         self.type = (type or self.default_type).lower()
         self.value = value
+        self.user_ids_with_ts = {}
         self.user_ids = set()
         self._recently_checked_db = recently_checked_db
 
         if user_ids:
-            self.add_user_ids(user_ids)
+            self.add_user_ids(user_ids, timestamps=timestamps)
 
         if self.domain is None:
             raise ValueError('domain should not be None')
@@ -134,21 +160,35 @@ class Stitch(object):
             raise ValueError('redis_client should not be None')
 
     def add_user_ids(self,
-                     user_ids):
+                     user_ids,
+                     timestamps=None):
 
         assert isinstance(user_ids, (list, set)), 'user_ids needs to be a list or set'
+        if timestamps:
+            assert isinstance(timestamps, list), 'timestamps should be a list'
+            assert len(timestamps) == len(user_ids), 'length of timestamps should be equal to length of user_ids'
+            assert all(isinstance(ts, (float, int)) for ts in timestamps), 'all timestamps should be floats or integers'
+        else:
+            now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
+            timestamps = [now for _ in user_ids]
+
+        self.user_ids_with_ts.update({user_id: max(ts, self.user_ids_with_ts.get(user_id, 0)) for user_id, ts in dict(zip(user_ids, timestamps)).items()})
         self.user_ids = self.user_ids.union(set(user_ids))
 
+    def delete_from_cache(self):
+        self.redis_client.delete(self.key)
+
     def to_cache(self):
-        now = pytz.utc.localize(datetime.datetime.utcnow()).timestamp()
         pipe = self.redis_client.pipeline()
-        pipe.zadd(self.key, {u: now for u in self.user_ids})
+        pipe.zadd(self.key, self.user_ids_with_ts)
         pipe.expire(self.key, self.user_ttl + self.user_data_ttl_margin)
         pipe.execute()
 
     def get(self):
-        user_ids = [u.decode() for u in self.redis_client.zrangebyscore(self.key, min='-inf', max='+inf')]
-        self.add_user_ids(user_ids)
+        result = self.redis_client.zrangebyscore(self.key, min='-inf', max='+inf', withscores=True)
+        user_ids = [u[0].decode() for u in result]
+        timestamps = [u[1] for u in result]
+        self.add_user_ids(user_ids, timestamps=timestamps)
         if not self.recently_checked_db:
             self.complement_from_db()
 
@@ -156,8 +196,14 @@ class Stitch(object):
                         user_ids):
 
         assert isinstance(user_ids, (list, set)), 'user_ids needs to be a list or set'
-        self.redis_client.zrem(*user_ids)
+        pipe = self.redis_client.pipeline()
+        pipe.zrem(self.key, *user_ids)
+        for user_id in user_ids:
+            pipe.zrem(self.stitch_pool_key(user_id=user_id), self.key)
+        # pipe.zrem(self.all_stitchpools_key(), )
+        pipe.execute()
         self.user_ids = self.user_ids - set(user_ids)
+        self.user_ids_with_ts = {k: v for k, v in self.user_ids_with_ts if not k in user_ids}
 
     def set_recent_db_check(self):
 
@@ -182,8 +228,8 @@ class Stitch(object):
             ]
         )
         data_dict = {k.decode(): cls.from_redis_data(key=k, data=dict(zip(v[::2], v[1::2])))
-                     for k,v in dict(zip(data[::2], data[1::2])).items()}
-        should_check_db = {k: v for k,v in data_dict.items() if not v.recently_checked_db}
+                     for k, v in dict(zip(data[::2], data[1::2])).items()}
+        should_check_db = {k: v for k, v in data_dict.items() if not v.recently_checked_db}
         checked_in_db = cls.complement_multiple_from_db(should_check_db)
         data_dict.update(checked_in_db)
 
@@ -193,10 +239,12 @@ class Stitch(object):
     @classmethod
     def complement_multiple_from_db(cls, should_check_dict):
 
-        stitch_models = list(cls.django_stitch_model.objects.filter(key__in=list(should_check_dict.keys())))
+        model = cls.usable_django_model()
+        stitch_models = list(model.objects.filter(key__in=list(should_check_dict.keys())))
         for stitch_model in stitch_models:
             should_check_dict[stitch_model.key] = should_check_dict[stitch_model.key].add_user_ids(
-                [str(u) for u in stitch_model.user_ids]
+                [str(u) for u in stitch_model.user_ids],
+                [float(t) for t in stitch_model.timestamps]
             )
             should_check_dict[stitch_model.key].to_cache()
             should_check_dict[stitch_model.key].set_recent_db_check()
@@ -204,14 +252,16 @@ class Stitch(object):
         return should_check_dict
 
     def complement_from_db(self):
+        model = self.usable_django_model()
         try:
-            stitch_model = self.django_stitch_model.objects.get(key=self.key)
+            stitch_model = model.objects.get(key=self.key)
             user_ids = [str(u) for u in stitch_model.user_ids]
-            self.add_user_ids(user_ids)
+            timestamps = [float(ts) for ts in stitch_model.timestamps]
+            self.add_user_ids(user_ids, timestamps=timestamps)
             if len(user_ids) > 0:
                 self.to_cache()
             self.set_recent_db_check()
-        except self.django_stitch_model.DoesNotExist:
+        except model.DoesNotExist:
             self.set_recent_db_check()
             return []
 
@@ -221,16 +271,53 @@ class Stitch(object):
             'type': self.type
         }
 
+    def to_db(self):
+
+        # first call complement_from_db!
+
+        if len(self.user_ids_with_ts) == 0:
+            raise Exception('Cannot store empty user_ids')
+        max_tries = 3
+        _tmp = list(self.user_ids_with_ts.items())
+        user_ids = [u[0] for u in _tmp]
+        timestamps = [u[1] for u in _tmp]
+
+        for i in range(max_tries):
+            try:
+                self.usable_django_model().objects.update_or_create(
+                    key=self.key,
+                    defaults={'user_ids': user_ids, 'timestamps': timestamps})
+                return
+            except Exception as e:
+                if 'OperationalError' in e.__class__.__name__:
+                    error = e
+                    time.sleep(1 + random.random() / 10)
+            raise error
+
+
     @classmethod
     def from_redis_data(cls,
                         key,
                         data):
 
         type, value, domain = cls.parse_key(key)
-        user_ids = [u.decode() for u in (data[b'user_ids'] or [])]
+        user_data_with_ts = (data[b'user_ids'] or [])
+        user_ids_with_timestamps = [(k.decode(), float(v)) for k,v in dict(zip(user_data_with_ts[::2], user_data_with_ts[1::2])).items()]
+        # user_ids_with_timestamps = [u.decode() for u in (data[b'user_ids'] or [])]
         checked_db = bool(data[b'recent_check'])
-        return cls(value=value, type=type, user_ids=user_ids, recently_checked_db=checked_db)
+        user_ids = [o[0] for o in user_ids_with_timestamps]
+        timestamps = [o[1] for o in user_ids_with_timestamps]
+        return cls(value=value, type=type, user_ids=user_ids, timestamps=timestamps, recently_checked_db=checked_db)
 
+    @classmethod
+    def usable_django_model(cls):
+        if cls.django_stitch_model:
+            if all(cls.django_connection_tuple):
+                if cls.django_connection_tuple[0].connection and not cls.django_connection_tuple[0].is_usable():
+                    del cls.django_connection_tuple[1]._connections.default
+                return cls.django_stitch_model
+            else:
+                return cls.django_stitch_model
 
     @classmethod
     def parse_key(cls, key):
@@ -240,6 +327,17 @@ class Stitch(object):
         type, value, domain = split_key[2], ':'.join(split_key[3:-1]), split_key[-1]
         return type, value, domain
 
+    @classmethod
+    def stitch_pool_key(cls,
+                        user_id):
+        # this is also defined on the stitch pool class in python-grains!
+        return f'st:u:p:{cls.domain}:{user_id}'
+
+    @classmethod
+    def all_stitchpools_key(cls):
+        # this is also defined on the stitch pool class in python-grains!
+        return f'st:u:p:all:{cls.domain}'
+
     @property
     def key(self):
         return self._key(type=self.type, value=self.value)
@@ -254,4 +352,12 @@ class Stitch(object):
 
     @classmethod
     def _recently_checked_db_key(cls, type, value):
-        return cls._key(type=type, value=value) + cls.recently_checked_db_postfix
\ No newline at end of file
+        return cls._key(type=type, value=value) + cls.recently_checked_db_postfix
+
+    @property
+    def all_stitches_key(self):
+        return self._all_stitches_key()
+
+    @classmethod
+    def _all_stitches_key(cls):
+        return f'st:u:all:{cls.domain}'
